Require Import ssreflect eqtype ssrbool ssrnat ssrfun seq.
Require Import heaps rels hprop domains stmod stcont stlog.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(********************************)
(* A file with random examples. *)
(********************************)

(* let's try some loops *)
(* first the infinite loop *)

Program 
Definition diverge0 A : ST [::] A := 
  Fix id : STsec [::] A (fun i => True, fun y i m => False).
Next Obligation. by apply: cont_id. Qed.

Implicit Arguments diverge0 [A].

Program
Definition diverge As A : 
  STsec As A (pre_lam (fun g i => True), 
              post_lam (fun g1 g2 y i m => False)) :=
  Do (diverge0 @@ (fun_lam (fun g => tt))).
Next Obligation.
split=>//=; first by move=>a i _; sauto; apply: sval_do.
move=>a1 a2 i1 i2 _ _ _ _. 
by vauto; apply: val_do=>//=.
Qed.

Implicit Arguments diverge [As A].

(* computing factorial *)

Fixpoint fact (n : nat) : nat := 
  if n is n'.+1 then n * fact n' else 1.

Definition factT := 
  STsec [:: nat:Type] nat
    (fun n i => True, 
     fun n y i m => i = m /\ y.1 = fact n.1 /\ y.2 = fact n.2).

Program Definition factorial := 
  Fix (fun loop : factT => 
         Ifbool (fun n => n == 0) then 
           ret @@ (fun n => [1])
         else
           hide loop @@ (fun n => [n - 1]);;
           ret @@ (fun x n => [n * x]) 
         with [:: nat:Type], nat, 
           (fun n i => True,
           fun n y i m => i = m /\ y.1 = fact n.1 /\ y.2 = fact n.2)).
Next Obligation.
split=>[[n][]|[n1][][n2][]] /=.
- move=>i _; case: n=>[|n] /=; sauto.
  by apply: sbnd_do=>//= y m _; sauto. 
move=>i1 i2 _ _ _ _; case: n1; case: n2=>/=.
- by vauto.
- move=>n2; vauto; lauto; apply: bnd_doV=>//= y m [[-> _]][-> _].
  by lauto=>_ _ _; rewrite subn1.
- move=>n1; vauto; apply: bnd_doL=>// y m [[-> _]] /= [-> _].
  by vauto; rewrite subn1.
move=>n1 n2; vauto; apply: bnd_do=>//= y1 y2 m1 m2 [[-> ->]][-> ->].
by vauto; rewrite !subn1.
Qed.
Next Obligation.
apply: if_bool_cont2; first by apply: const_stcont.
apply: bind_cont; last by apply: const_stcont.
by apply: inst_cont; apply: st_cont; apply: cont_id.
Qed.

(* ok, reasoning about continuity works *)
(* more loops will be given in other files; check llist *)
(* now we move to information flow properties *)


(* the example from Simonet *)
(* shows nested conditionals. in the example below, *)
(* we can specify that the output of the function *)
(* is dependent only on x and z, not on y *)
Program 
Definition simonet (a b c d : bool) :
  STsec [:: bool; bool; bool] bool
    (fun x y z i => True, 
     fun xx yy zz rr ii mm => rr.1 = (a == d) || (b == d) || xx.1 || zz.1 /\
                              rr.2 = (a == d) || (b == d) || xx.2 || zz.2) :=
  Do ((Ifbool (fun x y z => x) then 
          Ifbool (fun x y z => y) then 
             ret @@ (fun x y z => [a]) else ret @@ (fun x y z => [b])
          with [:: bool:Type; bool:Type; bool:Type], bool,
            (fun x y z i => True,
             fun x y z r i m => 
               r = (if y.1 then a else b, if y.2 then a else b))
        else 
          Ifbool (fun x y z => z) then 
            ret @@ (fun x y z => [a]) else ret @@ (fun x y z => [d])
          with [:: bool:Type; bool:Type; bool:Type], bool,
            (fun x y z i => True,
             fun x y z r i m => 
               r = (if z.1 then a else d, if z.2 then a else d))
        with [:: bool:Type; bool:Type; bool:Type], bool,
        (fun x y z i => True,
         fun x y z r i m => r = (if x.1 then if y.1 then a else b else 
                                 if z.1 then a else d, 
                                 if x.2 then if y.2 then a else b else 
                                 if z.2 then a else d)));;
        Ifbool (fun t x y z => (t == a) || (t == b)) then 
           ret @@ (fun t x y z => [true])
        else ret @@ (fun t x y z => [false])
        with [:: bool:Type; bool:Type; bool:Type; bool:Type], bool,
        (fun t x y z i => True,
         fun tt xx yy zz rr ii mm => 
           tt = (if xx.1 then if yy.1 then a else b else 
                              if zz.1 then a else d, 
                 if xx.2 then if yy.2 then a else b else 
                              if zz.2 then a else d) -> 
          if (a == d) || (b == d) then rr = (true, true) 
          else rr = (xx.1 || zz.1, xx.2 || zz.2))). 
Next Obligation.
split=>[[x][y][z][]|[x1][y1][z1][][x2][y2][z2][]] /=.
- by move=>i _; case: y; sauto.
by move=>i1 i2 _ _ _ _; case: y1; case: y2; vauto.
Qed.
Next Obligation.
split=>[[x][y][z][]|[x1][y1][z1][][x2][y2][z2][]] /=.
- by move=>i _; case: z; sauto.
by move=>i1 i2 _ _ _ _; case: z1; case: z2; vauto.
Qed.
Next Obligation.
split=>[[x][y][z][]|[x1][y1][z1][][x2][y2][z2][]] /=.
- by move=>i _; case: x; apply: sval_do.
move=>i1 i2 _ _ _ _. 
by case: x1; case: x2; first by [apply: val_do];
   apply: val_doL=>//= [r1] m1 [-> _];
   apply: val_doV=>//= [r2] m2 [-> _].
Qed.
Next Obligation.
rewrite !gh; split=>[[t][x][y][z][]|[t1][x1][y1][z1][][t2][x2][y2][z2][]] /=.
- by move=>i _; case: ifP=>_; sauto.
move=>i1 i2 _ [-> ->] _ _ _ _. 
case: ifP=>H1; vauto; case: ifP=>H2; vauto=>_ _; case: ifP=>E //.
- case: x1 H1; case: x2 H2; case: z1; case: z2=>//=;
  by rewrite 2!(eq_sym d) E.
- case: x2 H2; first by case: y2=>//=; rewrite !eq_refl 1?orbT.
  by case: z2=>//=; rewrite 2?(eq_sym d) ?E ?eq_refl.
- case: x2 H2; first by case: y2=>//=; rewrite eq_refl 1?orbT.
  case: z2; first by rewrite eq_refl.
  by case: x1 H1; case: z1=>//= ->.
- case: x1 H1; first by case: y1; rewrite eq_refl 1?orbT.
  case: z1; first by rewrite  eq_refl.
  by rewrite 2!(eq_sym d) E.
- case: x1 H1; first by case: y1; rewrite eq_refl 1?orbT.
  case: z1; first by rewrite eq_refl.
  by case: x2 H2=>//; case: z2=>// ->.
- case: x1 H1; first by case: y1; rewrite eq_refl 1?orbT.
  case: z1; first by rewrite eq_refl.
  by rewrite 2!(eq_sym d) E.
case: x1 H1; first by case: y1; rewrite eq_refl 1?orbT.
case: z1; first by rewrite eq_refl.
case: x2 H2; first by case: y2; rewrite eq_refl 1?orbT.
by case: z2=>//; rewrite eq_refl.
Qed.
Next Obligation.
split=>[[x][y][z][]|[x1][y1][z1][][x2][y2][z2][]] /=.
- move=>i _. 
  sauto; apply: sbnd_do=>//= y1 m1 [-> _].
  by sauto; apply: sval_do.
move=>i1 i2 _ _ _ _.
vauto; apply: bnd_do=>//= y3 y4 m1 m2 [-> ->].
vauto; apply: val_do=>//= y5 y6 m3 m4.
move/(_ (erefl _)).
by case: ifP=>E [-> ->].
Qed.

Program
Definition silly0 : 
  STsec [:: bool:Type] ptr
    (fun h i => True,
     fun hh yy ii mm => 
         mm.1 = ii.1 :+ yy.1 :-> 2 /\ 
         mm.2 = ii.2 :+ yy.2 :->2 /\ 
         (ii.1 =~ ii.2 -> yy.1 = yy.2)) :=
  Do (Ifbool (fun h => h) then 
         lalloc @@ (fun h => [2]);;
         ret @@ (fun z h => [z])
       else 
         halloc @@ (fun h => [1]);;
         lalloc @@ (fun y h => [2]);;
         dealloc @@ (fun z y h => [y]);;
         ret @@ (fun _ z y h => [z]) 
       with [:: bool:Type], ptr,
         (fun h i => True,
          fun hh yy ii mm => 
             mm.1 = ii.1 :+ yy.1 :-> 2 /\ 
             mm.2 = ii.2 :+ yy.2 :->2 /\ 
             (ii.1 =~ ii.2 -> yy.1 = yy.2))).
Next Obligation.
split=>[[h][]|[h1][][h2][]] /=.
- move=>i _.  
  case: h; first by sauto=>z; sauto.
  by sauto=>y; sauto=>z; sauto.
move=>i1 i2 _ _ _ _. 
case: h1; case: h2.
- vauto=>z1 z2; vauto=>_ _ L _ /=.
  by rewrite (unC i1) (unC i2); do !split=>//; case/L.
- apply/vrfC. 
  lauto=>y; apply: vrf2D=>D _; vauto=>z1 z2; lauto=>_ _ L H _ /=.
  rewrite (unC i1) (unC i2); do !split=>//; move=>E.
  have T: y :-> 1 :+ i2 =~ i1 by rewrite highPtUn // low_sym.
  by case: (L T)=>->.
- lauto=>y; apply: vrf2D=>D _; vauto=>z1 z2; lauto=>_ _ L H _ /=.
  rewrite (unC i1) (unC i2); do !split=>//; move=>E.
  have T : y :-> 1 :+ i1 =~ i2 by rewrite highPtUn.
  by case: (L T).
vauto=>y1 y2; apply: vrf2D=>D1 D2; vauto=>z1 z2; vauto=>_ _ L H1 H2 _.
rewrite (unC i1) (unC i2); do !split=>//; move=>E.
have T: y1 :-> 1 :+ i1 =~ y2 :-> 1 :+ i2 by apply: highPtUn2.
by case: (L T).
Qed.
Next Obligation.
split=>[[a][]|[a1][][a2][]] /=; first by move=>i _; apply: sval_do.
by move=>i1 i2 _ _ _ _; apply: val_do.
Qed.

Fixpoint public (ls : seq ptr) : Pred (d heap) :=
  if ls is l::ls then 
    fun i => look l i.1 = look l i.2 /\ public ls i
  else PredT.

Lemma public_refl ls i : public ls (i, i).
Proof. by elim: ls. Qed.

Hint Resolve public_refl. 

(* example from Terauchi and Aiken *)
(* demonstrates handling of implicit flows *)

Program
Definition silly (x y z h l : ptr) :
  STsec [::] unit
    (fun i => 
       exists f, exists u:nat, exists v:nat, 
       exists w:nat, exists b:bool, exists c:nat, 
         i = x :-> u :+ y :-> v :+ z :-> w :+ h :-> b :+ l :-> c :+ f, 
     fun r i m => public [:: l; y] i -> public [:: l; y] m) :=
  Do (write @ [z, 1];;
      read bool @@ (fun _ => [h]);;
      Ifbool (fun b _ => b) then 
        write @@ (fun b _ => [x, 1])
        else ret @@ (fun b _ => [tt]) 
      with [:: bool:Type; unit], unit,
           (fun _ _ i => exists u:nat, exists h, i = x :-> u :+ h, 
            fun b _ y i m=> 
              forall (u : d nat) h, 
                i.1 = x :-> u.1 :+ h.1 -> 
                i.2 = x :-> u.2 :+ h.2 ->
                (m.1 = if b.1 then x :-> 1 :+ h.1 else i.1) /\ 
                (m.2 = if b.2 then x :-> 1 :+ h.2 else i.2));;
      Ifbool (fun _ b _ => ~~ b) then 
        read nat @@ (fun _ _ _ => [z]);;
        write @@ (fun w _ _ _ => [x, w])
      else ret @@ (fun _ _ _ => [tt]) 
      with [:: unit:Type; bool; unit], unit,
           (fun _ _ _ i => exists u:nat, exists h, i = z :-> 1 :+ x :-> u :+ h, 
            fun _ b _ y i m=> 
              forall (u : d nat) h, 
                i.1 = z :-> 1 :+ x :-> u.1 :+ h.1 -> 
                i.2 = z :-> 1 :+ x :-> u.2 :+ h.2 ->
                (m.1 = if ~~ b.1 then z :-> 1 :+ x :-> 1 :+ h.1 else i.1) /\ 
                (m.2 = if ~~ b.2 then z :-> 1 :+ x :-> 1 :+ h.2 else i.2));;
      read nat @@ (fun _ _ b _ => [x]);;
      read nat @@ (fun u _ _ b _ => [y]);;
      write @@ (fun v u _ _ b _ => [l, u + v]);;
      ret @@ (fun _ v u _ _ b _ => [tt])).
Next Obligation.
split=>[[b][][][]|[b1][][][][b2][][][]] /=.
- by move=>i [u][r] ->; case: b; sauto.
move=>i1 i2 [u1][h1] -> [u2][h2] -> D1 D2.
case: b1; case: b2; lauto=>_ _ _ u j;
by case/(cancel D1)=>_ _ ->; case/(cancel D2)=>_ _ ->.
Qed.
Next Obligation.
split=>[[][][b][][][]|[][][b1][][][][][][b2][][][]] /=.
- by move=>i [u][r] ->; rewrite -!unA; case: b=>/=; sauto.
move=>i1 i2 [u1][h1] -> [u2][h2] ->; rewrite -!unA => D1 D2.
case: b1=>/=; case: b2=>/=; lauto=>_ _ _ v j E1 E2 //;
rewrite -!(pull x) -!unA in D1 D2 *;
move: E1; heap_cancel=>// _ -> _ _ //;
by move: E2; heap_cancel=>// _ -> _ _.
Qed.
Next Obligation.
rewrite !gh; split=>[[]|[][]] /=.
- move=>i [f][u][v][w][b][c] ->; rewrite -!unA; sauto.   
  apply: sbnd_do=>/= [|[] m]; rewrite -!(pull x); first by hhauto.
  case/(_ (u, u) (_, _) (erefl _) (erefl _))=>/= _ ->{m}.
  case: b; sauto;
  by apply: sbnd_do=>/= [|[] m]; rewrite -!(pull z) unA; first by [hhauto];
     case/(_ (_, _) (_, _) (erefl _) (erefl _))=>/= _ ->{m}; 
     rewrite -unA; sauto.
move=>i1 i2 _ [H1][H2] _.
move=>[f1][u1][v1][w1][b1][c1] E1 [f2][u2][v2][w2][b2][c2] E2 D1 D2.
move: H1 H2; rewrite {i1}E1 {i2}E2 -!unA in D1 D2 *.
rewrite -!(pull l) in D1 D2 *; rewrite !lookPtUn //.
rewrite -!(pull y) in D1 D2 *; rewrite !lookPtUn //.
do 2![move/dyn_inj=>->]; vauto=>{D1 D2}.
apply: bnd_do=>/=[|[][] m1 m2]; rewrite -!(pull x); first by hhauto.
case/(_ (_, _) (_, _) (erefl _) (erefl _))=>->->{m1 m2}.
case: b1; case: b2=>/=; vauto;
(apply: bnd_do=>/= [|[][] m1 m2]; 
 rewrite -!(pull z) !(unA (z :-> _)); first by [hhauto];
 case/(_ (_, _) (_, _) (erefl _) (erefl _))=>->->{m1 m2};
 rewrite -!unA);
vauto=>/= D1 D2; rewrite !lookPtUn //; 
by rewrite -!(pull y) in D1 D2 *; rewrite !lookPtUn.
Qed.

(* locations can be private, but their contents can be public *)
Program
Definition silly2 : 
  STsec [::ptr; ptr; ptr; ptr; ptr] unit 
    (fun x y z h l i => 
       exists f, exists u:nat, exists v:nat, 
       exists w:nat, exists b:bool, exists c:nat, 
         i = x :-> u :+ y :-> v :+ z :-> w :+ h :-> b :+ l :-> c :+ f, 
       fun x y z h l r i m => 
         look l.1 i.1 = look l.2 i.2 -> look y.1 i.1 = look y.2 i.2 -> 
         look l.1 m.1 = look l.2 m.2 /\ look y.1 m.1 = look y.2 m.2) :=
 Do (write @@ (fun x y z h l => [z, 1]);;
     read bool @@ (fun _ x y z h l => [h]);;
     Ifbool (fun b _ x y z h l => b) then 
       write @@ (fun b _ x y z h l => [x, 1])
       else ret @@ (fun b _ x y z h l => [tt]) 
     with [:: bool:Type; unit; ptr; ptr; ptr; ptr; ptr], unit,
       (fun _ _ x y z h l i => exists u:nat, exists h, i = x :-> u :+ h, 
        fun b _ x y z h l y i m=> 
          forall (u : d nat) h, 
            i.1 = x.1 :-> u.1 :+ h.1 -> 
            i.2 = x.2 :-> u.2 :+ h.2 ->
            (m.1 = if b.1 then x.1 :-> 1 :+ h.1 else i.1) /\ 
            (m.2 = if b.2 then x.2 :-> 1 :+ h.2 else i.2));;
     Ifbool (fun _ b _ x y z h l => ~~ b) then 
       read nat @@ (fun _ _ _ x y z h l => [z]);;
       write @@ (fun w _ _ _ x y z h l => [x, w])
     else ret @@ (fun _ _ _ x y z h l => [tt]) 
     with [:: unit:Type; bool; unit; ptr; ptr; ptr; ptr; ptr], unit,
          (fun _ _ _ x y z h l i => 
             exists u:nat, exists h, i = z :-> 1 :+ x :-> u :+ h, 
           fun _ b _ x y z h l y i m=> 
             forall (u : d nat) h, 
               i.1 = z.1 :-> 1 :+ x.1 :-> u.1 :+ h.1 -> 
               i.2 = z.2 :-> 1 :+ x.2 :-> u.2 :+ h.2 ->
               (m.1 = if ~~ b.1 then z.1 :-> 1 :+ x.1 :-> 1 :+ h.1 else i.1) /\ 
               (m.2 = if ~~ b.2 then z.2 :-> 1 :+ x.2 :-> 1 :+ h.2 else i.2));;
     read nat @@ (fun _ _ b _ x y z h l => [x]);;
     read nat @@ (fun u _ _ b _ x y z h l => [y]);;
     write @@ (fun v u _ _ b _ x y z h l => [l, u + v]);;
     ret @@ (fun _ v u _ _ b _ x y z h l => [tt])).
Next Obligation.
split=>[[b][][][]|] /=. 
- by move=>x [y][z][h][l][] i [u][r] ->; case: b; sauto.
case=>b1 [][][x1][y1][z1][h1][l1][][b2][][][x2][y2][z2][h2][l2][] /=.
move=>i1 i2 [u1][r1] -> [u2][r2] -> D1 D2.
case: b1; case: b2; lauto=>_ _ _ u j;
by case/(cancel D1)=>_ _ ->; case/(cancel D2)=>_ _ ->.
Qed.
Next Obligation.
split=>[[][][b][][][]|] /=. 
- by move=>x [y][z][h][l][] i [u][r] ->; rewrite -!unA; case: b=>/=; sauto.
case=>[][][b1][][][x1][y1][z1][h1][l1][].
case=>[][][b2][][][x2][y2][z2][h2][l2][] /=.
move=>i1 i2 [u1][r1]  -> [u2][r2] ->; rewrite -!unA => D1 D2.
case: b1=>/=; case: b2=>/=; lauto=>_ _ _ v j E1 E2 //;
rewrite -!(pull x1) -!(pull x2) -!unA in D1 D2 E1 E2 *;
move: E1; heap_cancel=>// _ -> _ _ //;
by move: E2; heap_cancel=>// _ -> _ _.
Qed.
Next Obligation.
rewrite !gh; split.
- move=>[x][y][z][h][l][i][f][u][v][w][b][c] -> /=; rewrite -!unA; sauto.   
  apply: sbnd_do=>/= [|[] m]; rewrite -!(pull x); first by hhauto.
  case/(_ (u, u) (_, _) (erefl _) (erefl _))=>/= _ ->{m}.
  case: b; sauto;
  by apply: sbnd_do=>/= [|[] m]; rewrite -!(pull z) unA; first by [hhauto];
     case/(_ (_, _) (_, _) (erefl _) (erefl _))=>/= _ ->{m}; 
     rewrite -unA; sauto.
case=>[x1][y1][z1][h1][l1][][x2][y2][z2][h2][l2][] i1 i2 _ /= [H1 H2].
move=>[f1][u1][v1][w1][b1][c1][E1][f2][u2][v2][w2][b2][c2][E2] D1 D2.
move: H1 H2; rewrite {i1}E1 {i2}E2 -!unA in D1 D2 *.
rewrite -!(pull l1) -!(pull l2) in D1 D2 *; rewrite !lookPtUn //.
rewrite -!(pull y1) -!(pull y2) in D1 D2 *; rewrite !lookPtUn //.
do 2![move/dyn_inj=>->]; vauto=>{D1 D2}.
apply: bnd_do=>/=[|[][] m1 m2]; rewrite -!(pull x1) -!(pull x2); 
  first by hhauto.
case/(_ (_, _) (_, _) (erefl _) (erefl _))=>->->{m1 m2}.
case: b1; case: b2=>/=; vauto;
(apply: bnd_do=>/= [|[][] m1 m2]; 
 rewrite -!(pull z1) -!(pull z2) !(unA (z1 :-> _)) !(unA (z2 :-> _)); 
   first by [hhauto];
 case/(_ (_, _) (_, _) (erefl _) (erefl _))=>->->{m1 m2};
 rewrite -!unA);
vauto=>/= D1 D2; rewrite !lookPtUn //; 
by rewrite -!(pull y1) -!(pull y2) in D1 D2 *; rewrite !lookPtUn.
Qed.

(* Heap locations can be declared public not only by stating the
equality of their contents in the two heaps, but by stating something
that implies this equality. The program below exemplifies this; in the
spec we have a hypothesis that the sum of the two variables x and y
must be 0 in both heaps, hence implying that x and y must be equal to
0. since their values are equal (and equal to 0) in the input heaps,
both x and y must be public.  *)

Program 
Definition silly3 (x y : ptr) : 
   STsec [::] nat 
     (fun i => exists h, exists v:nat, exists w:nat, 
                 i = x :-> v :+ y :-> w :+ h,
      fun z i m => forall (h : d heap) (v w : d nat), 
        i.1 = x :-> v.1 :+ y :-> w.1 :+ h.1 -> 
        i.2 = x :-> v.2 :+ y :-> w.2 :+ h.2 -> 
        v.1 + w.1 = v.2 + w.2 -> 
        v.1 + w.1 = 0 -> m = i /\ z = (0, 0)) :=
   Do (read nat @ [x];;
       read nat @@ (fun v1 => [y]);;
       ret @@ (fun v2 v1 => [v1])).
Next Obligation.
rewrite !gh /=; split=>[[]|[][]]. 
- by move=>i [h][v][w] ->; rewrite -unA; sauto.
move=>i1 i2 [[[[]]]][f1 f2][v1 v2][w1 w2] /= [[[-> ->]]] E1 E2 _ _ _ _.
rewrite -!unA; vauto=>/= _ _.
rewrite E2 in E1; move/eqP: (esym E1); move/eqP: E2; rewrite !addn_eq0.
by do 2![case/andP; move/eqP=>-> _].
Qed.

(* another example with polynomials *)
Definition sgn (x : nat) : nat := if x > 0 then 1 else 0.

Program
Definition poly : 
  STsec [:: nat; nat; nat; nat] nat 
    (fun a b c x i => True,
     fun a b c x y i m => 
       sgn (a.1*x.1*x.1 + b.1*x.1 + c.1) = 
       sgn (a.2*x.2*x.2 + b.2*x.2 + c.2) -> y.1= y.2) :=
  Do (ret @@ (fun a b c x => [sgn (a*x*x + b*x + c) + 1])).
Next Obligation.
rewrite !gh; split=>/=; first by case=>a [b][c][x][] i _; sauto.
case=>a1 [b1][c1][x1][][a2][b2][c2][x2][] /=.
by move=>i1 i2 _ E _ _ _ _; vauto=>_ _; rewrite E.
Qed. 

(* A number of examples about granting access to locations, and hiding *)
(* The examples here are done in an "object-oriented" fashion *)
(* In the alice.v file, the examples are redone to match the paper *)

(* Access to a location can be controled by means of a type system in a *)
(* fairly standard way. To grant read access, but not write access to a *)
(* location, simply hide the location by means of type and predicate *)
(* abstraction, and export a function for reading this location in the *)
(* module interface. The following example does just that. It does not *)
(* have much to do with information leakage, but it illustrates other *)
(* features, such as access control via types. *)

Structure alice : Type :=
  Alice {shape : nat -> Pred heap; 
         get_secret : 
           STsec [::] nat 
             (fun i => exists h, exists v, i \In shape v # this h, 
              fun y i m => forall h v, 
                i.1 \In shape v.1 # this h.1 -> 
                  i.2 \In shape v.2 # this h.2 -> 
                m.1 \In shape v.1 # this h.1 /\ 
                m.2 \In shape v.2 # this h.2 /\ y = v)}.

Program
Definition new_alice (secret : nat) : 
  STsec [::] alice 
    (fun i => True, 
     fun y i m => m.1 \In shape y.1 secret # this i.1 /\ 
                  m.2 \In shape y.2 secret # this i.2) :=
  Do (halloc @ [secret];;
      ret @@ (fun x => [@Alice (fun n i => i = x :-> n)
                               (Do (read nat @ [x];; 
                                    ret @@ fun y => [y]))])). 
Next Obligation.
rewrite !gh; split=>[[]|[][]] /=.
- by move=>_ [h][v][_][_][->][[->]][<-]; sauto. 
move=>_ _ [[[]]][i1 i2][v1 v2] /= [[_]][_][->][[->]]
          [<-][_][_][->][[->]][<-] _ _ _ _. 
by vauto; do !split=>//; hhauto.
Qed.
Next Obligation.
split=>[[]|[][]] /=; first by move=>i _; sauto=>x; sauto. 
by move=>i1 i2 _ _ _ _; vauto=>x1 x2; vauto; hhauto.
Qed.

(* Now, someone trying to access alice's local store, can do so only via *)
(* a call to get_secret, as alice's shape predicate is abstracted. It is *)
(* important to understand how this abstraction came to be, as this *)
(* illustrates one of the main properties of the system, namely, the Do *)
(* construct. *)

(* Notice that the computation of new returns a fresh alice structure, *)
(* which is actually a dependent pair including get_secret, and a *)
(* predicate describing the shape of alice's local state. In principle, *)
(* if dependent pairs were all there was to this program, we would have *)
(* not actually achieved any information hiding, because one can project *)
(* out the components of a dependent pair, and hence access the *)
(* supposedly abstracted shape predicate. *)

(* But here, the construction of the alice structure occurs in a monadic *)
(* computation, in the scope of the Do construct. Because the logic and *)
(* the implementation do not expose the semantics of Do (notice that Do's *)
(* definition is deliberately made opaque in the stcont.v file), the only *)
(* knowledge that the clients of new will have about the returned alice *)
(* value is that it satisfies the Alice type. Nothing else is exported *)
(* about this value; in particular, clients will have to treat alice's *)
(* shape predicate as an un-interpreted function. As a consequence, the *)
(* clients can only use the get_secret function to read alice's secret, *)
(* but they can't do anything else, and in particular, they can't write *)
(* into alice's secret location because writing requires first deriving *)
(* out of this shape predicate that the written location is allocated. *)

(* For example, in the program and the proof below, one can try to unfold *)
(* the definition of shape, but that will not reveal any information *)
(* beyond alice's spec. *)

Program 
Definition do_the_read : 
  STsec [::] nat (PredT, fun y i m => y.1 = y.2) := 
  Do (new_alice 0;;
      apply get_secret @@ (fun a => [a])).
Next Obligation.
split=>[[]|[][]] /=.
- move=>i _; sauto.
  apply: sbnd_do=>[|y m] //= [[h1][h2]][[->]][/= H1] _ _.
  by apply: sval_do=>/=; do !rewrite instE /=; hhauto; eauto.
move=>i1 i2 _ _ _ _; vauto. 
apply: bnd_do=>// y1 y2 m1 m2 [/= H1 H2]; vauto. 
apply: val_doL=>/= [|v1 m3]; first by exists i1; exists 0.
case/(_ (erefl _) (i1, i1) (0, 0) H1 H1)=>/= H3 [_][-> _].
apply: val_doV=>[|v2 m4]=>/=; first by exists i2; exists 0. 
by case/(_ (erefl _) (i2, i2) (0, 0) H2 H2)=>/= H4 [_][-> _].
Qed. 

(* The above example illustrates that combining Do with dependent pairing *)
(* suffices to enforce predicate abstraction, which leads to enforcing *)
(* correct access control. We next give another important example of *)
(* using Do. *)

(* Notice that the type of do_the_read not only enforces access control, *)
(* but it also exposes the fact that alice's stored secret is equal to *)
(* the argument passed to new. In particular, the claimed secret is *)
(* actually clearly equal in both the runs of do_the_read, and is hence *)
(* public. *)

(* If we want to make alice's secret really secret, we have to hide the *)
(* later equality property upon creating instances of alice. This can be *)
(* done by using a different new function, as below, which hides the fact *)
(* that the stored secret actually equals the initially passed input.  *) 

Program 
Definition new_alice1 (secret : nat) : 
  STsec [::] alice 
    (PredT, 
     fun y i m => exists v, m.1 \In shape y.1 v.1 # this i.1 /\ 
                            m.2 \In shape y.2 v.2 # this i.2) := 
  Do (new_alice secret).
Next Obligation.
split=>[[]|[][]] /=; first by move=>i _; apply: sval_do.
move=>i1 i2 _ _ _ _ .
apply: val_do=>// y1 y2 m1 m2 [/= H1 H2] _ _ _.
by exists (secret, secret); hhauto.
Qed.

(* If alice is created by new1, clients accessing her secret via *)
(* get_secret, won't be able to prove that the obtained value is public. *)

Program
Definition do_the_read1 : 
  STsec [::] nat (PredT, fun y i m => y.1 = y.2) := 
  Do (new_alice1 0;; 
      apply get_secret @@ (fun a => [a])).
Next Obligation.
split=>[[] i _|[][] i1 i2 _ _ _ _] /=.
- sauto; apply: sbnd_do=>// y m [[v _]] /= [H _]; sauto. 
  by apply: sval_do=>/=; exists i; exists v.
vauto; apply: bnd_do=>// y1 y2 m1 m2 [[v1 v2]][/= H1 H2].
vauto; apply: val_doL=>[|z1 n1] /=; first by exists i1; exists v1.
case/(_ (erefl _) (i1, i1) (v1, v1) H1 H1)=>/= H3 [H4][-> _].
apply: val_doV=>[|z2 n2] /=; first by exists i2; exists v2.
case/(_ (erefl _) (i2, i2) (v2, v2) H2 H2)=>/= H5 [H6][-> _] _ _.
(* can't push the proof through; the spec is not valid *)
(* as expected, we have to admit *)
admit.
Qed.

(* Alternativelly, we can expose that alice's secret equals the input of *)
(* new, but then make this input be secret. Clients who have already *)
(* managed to declassify the input will simply propagate this knowledge *)
(* forward. *)

Program 
Definition new_alice2 :
  STsec [:: nat] alice 
    (fun s i => True, 
     fun s y i m => m.1 \In shape y.1 s.1 # this i.1 /\ 
                    m.2 \In shape y.2 s.2 # this i.2) :=
  Do (halloc @@ (fun s => [s]);;
      ret @@ (fun x s => [@Alice (fun n i => i = x :-> n)
                                 (Do (read nat @ [x];;
                                      ret @@ (fun y => [y])))])).
Next Obligation.
rewrite !gh; split=>[[] i|[][]] /=.
- by case=>h [v][_][_][->][[->]][<-]; sauto. 
move=>_ _ [[[]]][i1 i2][v1 v2][[_]][_][->][[->]][<-][_][_][->][[->]][<-]. 
by move=>_ _ _ _; vauto; hhauto.
Qed.
Next Obligation.
split=>[[s][] i _|[s1][][s2][] i1 i2 _ _ _ _] /=; first by sauto=>x; sauto. 
by vauto=>x1 x2; vauto; hhauto.
Qed.

(* Now, say alice wants to use somebody's program for managing secrets, *)
(* i.e., divide the secret by two. Alice wants to allow the managing *)
(* program to access the secret and modify it, but she wants to ensure *)
(* that her secret is not stolen. Let's say bob is a code provider, who *)
(* exports a secret-processing program to be run by instances of *)
(* alice. Bob wants to keep statistics on how many time his program has *)
(* been executed by alice. But, bob also wants to convince alice that his *)
(* program will not steal any of alices secret information. The next *)
(* example shows how these properties can be specified. *)

Structure bob : Type := 
  Bob {bshape : nat -> Pred heap;
       process : forall a, 
         STsec [::] nat 
           (fun i => exists h, exists v, exists k, 
                       i \In shape a v # bshape k # this h, 
            fun y i m => 
              forall h v k j s, 
                i.1 = j.1 :+ s.1 :+ h.1 -> j.1 \In shape a v.1 ->
                i.2 = j.2 :+ s.2 :+ h.2 -> j.2 \In shape a v.2 -> 
                s.1 \In bshape k.1 -> s.2 \In bshape k.2 ->
              exists n, exists t, 
                m.1 = n.1 :+ t.1 :+ h.1 /\ n.1 \In shape a v.1 /\
                m.2 = n.2 :+ t.2 :+ h.2 /\ n.2 \In shape a v.2 /\ 
                t.1 \In bshape k.1.+1 /\ t.2 \In bshape k.2.+1 /\
                y = (v.1./2, v.2./2) /\ 
                (i.1 =~ i.2 -> s.1 = s.2 -> t.1 = t.2))}. 
                     
Program 
Definition new_bob :
  STsec [::] bob 
    (fun i => True, 
     fun y i m => m.1 \In bshape y.1 0 # this i.1 /\ 
                  m.2 \In bshape y.2 0 # this i.2) :=
  Do (halloc @ [0];; 
      ret @@ (fun x => [@Bob (fun n i => i = x :-> n)
                             (fun a => 
                                Do(get_secret a;;
                                   read nat @@ (fun y => [x]);;
                                   write @@ (fun z y => [x, z.+1]);;
                                   ret @@ (fun _ z y => [y./2])))])).
Next Obligation. 
rewrite !gh; split=>[[]|[][]] /=.
- move=>_ [h][v][k][i][_][->][H][_][_][->][[->]][<-].
  have {H} H: i :+ (x :-> k :+ h) \In shape a v # this (x :-> k :+ h).
  - by hhauto. 
  sauto. 
  apply: sbnd_do=>[|y m] /=; first by exists (x :-> k :+ h); exists v.
  case/(_ (_, _) (_, _) H H)=>{H} /= [[{i} i]][_][->][H1][<-][_][-> _].
  by sauto.
move=>_ _ [[[[[[]]]]]] /= h v k [i1 i2][_ _] /= [[[[[->]]]]] H1 -> H2 [->][->].
move=>_ _ D _; vauto.  
apply: bnd_do=>[|y1 y2 m1 m2] /=.
- by split; econstructor; [exists v.1 | exists v.2]; rewrite -unA; hhauto.
have {H1 H2}[H1 H2] : 
  i1 :+ x :-> k.1 :+ h.1 \In shape a v.1 # this (x :-> k.1 :+ h.1) /\
  i2 :+ x :-> k.2 :+ h.2 \In shape a v.2 # this (x :-> k.2 :+ h.2).
- by split; rewrite -unA; hhauto.
case/(_ (_, _) (_, _) H1 H2)=>{H1 H2} /=.
case=>h1 [_][->][H1][<-][[h2]][_][->][H2][<-][-> ->]; vauto=>_ _.
exists (h1, h2); exists (x :-> k.1.+1, x :-> k.2.+1).
rewrite /= !unA -!(pull x); do !split=>//; move=>_.
by case/(pts_inj (defUnr (defUnl D)))=>->.
Qed.
Next Obligation.
split=>[[]|[][]] /=; first by move=>i _; sauto=>x; sauto.
by move=>i1 i2 _ _ _ _; vauto=>x1 x2; vauto; hhauto.
Qed.

(* If we were to leak some information from alice into bob's local state, *)
(* we wouldn't have been able to prove the spec. *)

(* Say, for example, that bob keeps his local counter not as one pointer, *)
(* but as two. The count it wants to return is the difference of the *)
(* two. if bob wants to steal some information from alice, he can do so *)
(* as follows, without violating the correctness spec (but he will *)
(* violate his information flow leakage spec). He can increment both her *)
(* counters with the value of alices secret. Because of the way bob *)
(* computes the count, the two values would cancel each other, so the *)
(* count will be kept correctly. However, after this has been done, *)
(* alice's secret is effectively written into bob's local state, and is *)
(* hence stolen. *)

Program 
Definition new_bob1 :
  STsec [::] bob 
    (fun i => True, 
     fun y i m => m.1 \In bshape y.1 0 # this i.1 /\ 
                  m.2 \In bshape y.2 0 # this i.2) :=
  Do (halloc @ [0];; 
      halloc @@ (fun x1 => [0]);;
      ret @@ (fun x2 x1 => 
             [@Bob (fun n i => 
                      exists c, i = x1 :-> c.1 :+ x2 :-> c.2 /\ 
                                n = c.1 - c.2 /\ c.2 <= c.1)
                   (fun a => 
                      Do (get_secret a;;
                          read nat @@ (fun y => [x1]);;
                          read nat @@ (fun z1 y => [x2]);;
                          write @@ (fun z2 z1 y => [x1, z1.+1 + y]);;
                          write @@ (fun _ z2 z1 y => [x2, z2 + y]);;
                          ret @@ (fun _ _ z2 z1 y => [y./2])))])).
Next Obligation. 
rewrite !gh; split=>[[]|[][]] /=.
- move=>_ [h][v][k][i][h2][->][H][_][_][->][[c]][->] _ [<-].
  have {H} H: i :+ (x1 :-> c.1 :+ x2 :-> c.2 :+ h)
                  \In shape a v # this (x1 :-> c.1 :+ x2 :-> c.2 :+ h)
    by hhauto. 
  sauto; apply: sbnd_do=>[|y m]. 
  - by exists (x1 :-> c.1 :+ x2 :-> c.2 :+ h); exists v.
  case/(_ (_, _) (_, _) H H)=>{H} /= [[{i} i]][_][->][H1][<-][_][-> _].
  by rewrite -!unA; sauto.
move=>j1 j2 [[[[[[]]]]]] /= h v [k1 k2][i1 i2][s1 s2] /= [[[[[->]]]]] H1 -> H2.
case=>c1 [->{s1 j1}][->{k1}] E1 [c2][->{s2 j2}][->{k2}] E2 _ _ D _.
vauto; apply: bnd_do=>[|y1 y2 m1 m2] /=.
- by split; econstructor; [exists v.1 | exists v.2]; rewrite -unA; hhauto.
have {H1 H2}[H1 H2] : 
  i1 :+ (x1 :-> c1.1 :+ x2 :-> c1.2) :+ h.1 \In 
     shape a v.1 # this ((x1 :-> c1.1 :+ x2 :-> c1.2) :+ h.1) /\
  i2 :+ (x1 :-> c2.1 :+ x2 :-> c2.2) :+ h.2 \In 
     shape a v.2 # this ((x1 :-> c2.1 :+ x2 :-> c2.2) :+ h.2).
- by split; rewrite -unA; hhauto.  
case/(_ (_, _) (_, _) H1 H2)=>{H1 H2} /=.
case=>h1 [_][->][H1][<-][[h2]][_][->][H2][<-][-> ->]. 
rewrite -!unA; vauto=>_ _.
exists (h1, h2).
exists (x1 :-> (c1.1.+1 + v.1) :+ x2 :-> (c1.2 + v.1),  
        x1 :-> (c2.1.+1 + v.2) :+ x2 :-> (c2.2 + v.2)).
rewrite /= -!unA -!(pull x2) -!(pull x1); do 4![split=>//].
split.
- exists (c1.1.+1 + v.1, c1.2 + v.1); split=>//.
  by rewrite subn_add2r leq_subS // leq_add2r ltnW.
split.
- exists (c2.1.+1 + v.2, c2.2 + v.2); split=>//.
  by rewrite subn_add2r leq_subS // leq_add2r ltnW.
split=>// L; move: (defUnr (defUnl D))=>D1; heap_cancel=>// _ _ -> ->. 
(* we have pushed the proof as far as it can go *)
(* clearly, this program steals alice's secret v  *)
(* so we can't prove that it's safe; we have to admit *)
admit.
Qed.
Next Obligation.
split=>[[]|[][]] /=; first by move=>i _; do !sauto=>?. 
move=>i1 i2 _ _ _ _; vauto=>x1 x2; vauto=>y1 y2; vauto=>/= _ _ H1 H2 H3 H4 _.
rewrite !unA -!(pull x1) -!(pull x2); split;
by econstructor; econstructor; do !split=>//; exists (0, 0).
Qed.

(* we can now execute bob's process over several alices, without worry *)
(* that we are leaking any information from alices' secrets *)

Program
Definition client : STsec [::] nat (PredT, fun y i m => True) := 
  Do (new_alice2 @ [1];;
      new_alice2 @@ (fun a1 => [2]);;
      new_bob @@ (fun a2 a1 => tt);;
      apply (fun b => apply (process b)) @@ (fun b a2 a1 => [b, a1]);;
      apply (fun b => apply (process b)) @@ (fun t1 b a2 a1 => [b, a2]);;
      ret @@ (fun t2 t1 b a2 a1 => [t1 + t2])).
Next Obligation.
split=>[[] i _|[][]] /=; last by move=>i1 i2 _ _ _ _; vauto; apply: bnd_do.
sauto; apply: sbnd_do=>//= a1 _ [[h1]][_][->][H1][<-] _.
sauto; apply: sbnd_do=>//= a2 _ [[h2]][_][->][H2][<-] _.
sauto; apply: sbnd_do=>//= c h3 [[h4]][_][->][H3][<-] _.
sauto; apply: sbnd_do=>[|y1 m1]. 
- by rewrite -!(unCA h1); econstructor; exists 1; exists 0; hhauto. 
rewrite -!(unCA h1). 
move/(_ (erefl _) (erefl _) (h2 :+ i, h2 :+ i) (1, 1) (0, 0) (h1, h1) (h4, h4)).
rewrite !unA; case/(_ (erefl _) H1 (erefl _) H1 H3 H3).
move=>[i4 w1][[i5 w2]] /= [->]{H1}[H1] [E][_]{H3}[H3][_][[-> _]] L.
sauto; apply: sbnd_do=>[|y2 m2]; last by move=>_; sauto. 
exists (i4 :+ i); exists 2; exists 1. 
by rewrite -!(unCA h2) -unA -(unCA i5); hhauto.
Qed.
